@@ -1,6 +1,6 @@
 # -*- coding: utf-8 -*-
 # Copyright (c) 2009-2011, 2013-2014 LOGILAB S.A. (Paris, FRANCE) <contact@logilab.fr>
-# Copyright (c) 2014-2018 Claudiu Popa <pcmanticore@gmail.com>
+# Copyright (c) 2014-2020 Claudiu Popa <pcmanticore@gmail.com>
 # Copyright (c) 2014 Google, Inc.
 # Copyright (c) 2014 Eevee (Alex Munroe) <amunroe@yelp.com>
 # Copyright (c) 2015-2016 Ceridwen <ceridwenv@gmail.com>
@@ -9,9 +9,14 @@
 # Copyright (c) 2017-2018 Ashley Whetter <ashley@awhetter.co.uk>
 # Copyright (c) 2017 Łukasz Rogalski <rogalski.91@gmail.com>
 # Copyright (c) 2017 rr- <rr-@sakuya.pl>
-# Copyright (c) 2018 Bryce Guinta <bryce.paul.guinta@gmail.com>
 # Copyright (c) 2018 Nick Drozd <nicholasdrozd@gmail.com>
+# Copyright (c) 2018 Ville Skyttä <ville.skytta@iki.fi>
+# Copyright (c) 2018 Bryce Guinta <bryce.paul.guinta@gmail.com>
 # Copyright (c) 2018 HoverHell <hoverhell@gmail.com>
+# Copyright (c) 2019 Hugo van Kemenade <hugovk@users.noreply.github.com>
+# Copyright (c) 2020 hippo91 <guillaume.peillex@gmail.com>
+# Copyright (c) 2020 Vilnis Termanis <vilnis.termanis@iotics.com>
+# Copyright (c) 2020 Ram Rachum <ram@rachum.com>
 
 # Licensed under the LGPL: https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html
 # For details: https://github.com/PyCQA/astroid/blob/master/COPYING.LESSER
@@ -314,21 +319,28 @@ nodes.AssignAttr.assigned_stmts = assend_assigned_stmts
 def _arguments_infer_argname(self, name, context):
     # arguments information may be missing, in which case we can't do anything
     # more
-    if not (self.args or self.vararg or self.kwarg):
+    if not (self.arguments or self.vararg or self.kwarg):
         yield util.Uninferable
         return
+
+    functype = self.parent.type
     # first argument of instance/class method
-    if self.args and getattr(self.args[0], "name", None) == name:
-        functype = self.parent.type
+    if (
+        self.arguments
+        and getattr(self.arguments[0], "name", None) == name
+        and functype != "staticmethod"
+    ):
         cls = self.parent.parent.scope()
         is_metaclass = isinstance(cls, nodes.ClassDef) and cls.type == "metaclass"
         # If this is a metaclass, then the first argument will always
         # be the class, not an instance.
+        if context.boundnode and isinstance(context.boundnode, bases.Instance):
+            cls = context.boundnode._proxied
         if is_metaclass or functype == "classmethod":
             yield cls
             return
         if functype == "method":
-            yield bases.Instance(cls)
+            yield cls.instantiate_class()
             return
 
     if context and context.callcontext:
@@ -339,6 +351,9 @@ def _arguments_infer_argname(self, name, context):
     if name == self.vararg:
         vararg = nodes.const_factory(())
         vararg.parent = self
+        if not self.arguments and self.parent.name == "__init__":
+            cls = self.parent.parent.scope()
+            vararg.elts = [cls.instantiate_class()]
         yield vararg
         return
     if name == self.kwarg:
@@ -362,7 +377,7 @@ def arguments_assigned_stmts(self, node=None, context=None, assign_path=None):
         callcontext = context.callcontext
         context = contextmod.copy_context(context)
         context.callcontext = None
-        args = arguments.CallSite(callcontext)
+        args = arguments.CallSite(callcontext, context=context)
         return args.infer_argument(self.parent, node.name, context)
     return _arguments_infer_argname(self, node.name, context)
 
@@ -459,7 +474,7 @@ def _infer_context_manager(self, mgr, context):
             )
 
         for decorator_node in func.decorators.nodes:
-            decorator = next(decorator_node.infer(context))
+            decorator = next(decorator_node.infer(context=context))
             if isinstance(decorator, nodes.FunctionDef):
                 if decorator.qname() == _CONTEXTLIB_MGR:
                     break
@@ -486,8 +501,8 @@ def _infer_context_manager(self, mgr, context):
     elif isinstance(inferred, bases.Instance):
         try:
             enter = next(inferred.igetattr("__enter__", context=context))
-        except (exceptions.InferenceError, exceptions.AttributeInferenceError):
-            raise exceptions.InferenceError(node=inferred)
+        except (exceptions.InferenceError, exceptions.AttributeInferenceError) as exc:
+            raise exceptions.InferenceError(node=inferred) from exc
         if not isinstance(enter, bases.BoundMethod):
             raise exceptions.InferenceError(node=enter)
         yield from enter.infer_call_result(self, context)
@@ -672,11 +687,12 @@ def starred_assigned_stmts(self, node=None, context=None, assign_path=None):
                         break
                     elts.pop()
                     continue
-                # We're done
+
+                # We're done unpacking.
                 packed = nodes.List(
                     ctx=Store, parent=self, lineno=lhs.lineno, col_offset=lhs.col_offset
                 )
-                packed.postinit(elts=elts)
+                packed.postinit(elts=list(elts))
                 yield packed
                 break
 
